@@ -35,6 +35,7 @@
 
 #include "ld.hpp"
 #include "Snapshot.h"
+#include "MachOFileAbstraction.hpp"
 
 extern void throwf (const char* format, ...) __attribute__ ((noreturn,format(printf, 1, 2)));
 extern void warning(const char* format, ...) __attribute__((format(printf, 1, 2)));
@@ -86,6 +87,46 @@ public:
 	enum UUIDMode { kUUIDNone, kUUIDRandom, kUUIDContent };
 	enum LocalSymbolHandling { kLocalSymbolsAll, kLocalSymbolsNone, kLocalSymbolsSelectiveInclude, kLocalSymbolsSelectiveExclude };
 	enum DebugInfoStripping { kDebugInfoNone, kDebugInfoMinimal, kDebugInfoFull };
+#if SUPPORT_APPLE_TV
+	enum Platform { kPlatformUnknown, kPlatformOSX, kPlatformiOS, kPlatformWatchOS, kPlatform_tvOS };
+#else
+	enum Platform { kPlatformUnknown, kPlatformOSX, kPlatformiOS, kPlatformWatchOS };
+#endif
+
+	static Platform platformForLoadCommand(uint32_t lc) {
+		switch (lc) {
+			case LC_VERSION_MIN_MACOSX:
+				return kPlatformOSX;
+			case LC_VERSION_MIN_IPHONEOS:
+				return kPlatformiOS;
+			case LC_VERSION_MIN_WATCHOS:
+				return kPlatformWatchOS;
+		#if SUPPORT_APPLE_TV
+			case LC_VERSION_MIN_TVOS:
+				return kPlatform_tvOS;
+		#endif
+		}
+		assert(!lc && "unknown LC_VERSION_MIN load command");
+		return kPlatformUnknown;
+	}
+
+	static const char* platformName(Platform platform) {
+		switch (platform) {
+			case kPlatformOSX:
+				return "OSX";
+			case kPlatformiOS:
+				return "iOS";
+			case kPlatformWatchOS:
+				return "watchOS";
+		#if SUPPORT_APPLE_TV
+			case kPlatform_tvOS:
+				return "tvOS";
+		#endif
+			case kPlatformUnknown:
+			default:
+				return "(unknown)";
+		}
+	}
 
 	class FileInfo {
     public:
@@ -208,7 +249,7 @@ public:
 	bool						allowSubArchitectureMismatches() const { return fAllowCpuSubtypeMismatches; }
 	bool						forceCpuSubtypeAll() const { return fForceSubtypeAll; }
 	const char*					architectureName() const { return fArchitectureName; }
-	void						setArchitecture(cpu_type_t, cpu_subtype_t subtype);
+	void						setArchitecture(cpu_type_t, cpu_subtype_t subtype, Options::Platform platform);
 	bool						archSupportsThumb2() const { return fArchSupportsThumb2; }
 	OutputKind					outputKind() const { return fOutputKind; }
 	bool						prebind() const { return fPrebind; }
@@ -233,6 +274,7 @@ public:
 	bool						allGlobalsAreDeadStripRoots() const;
 	bool						shouldExport(const char*) const;
 	bool						shouldReExport(const char*) const;
+	std::vector<const char*>	exportsData() const;
 	bool						ignoreOtherArchInputFiles() const { return fIgnoreOtherArchFiles; }
 	bool						traceDylibs() const	{ return fTraceDylibs; }
 	bool						traceArchives() const { return fTraceArchives; }
@@ -240,7 +282,10 @@ public:
 	UndefinedTreatment			undefinedTreatment() const { return fUndefinedTreatment; }
 	ld::MacVersionMin			macosxVersionMin() const { return fMacVersionMin; }
 	ld::IOSVersionMin			iOSVersionMin() const { return fIOSVersionMin; }
+	ld::WatchOSVersionMin		watchOSVersionMin() const { return fWatchOSVersionMin; }
+	uint32_t					minOSversion() const;
 	bool						minOS(ld::MacVersionMin mac, ld::IOSVersionMin iPhoneOS);
+	bool						min_iOS(ld::IOSVersionMin requirediOSMin);
 	bool						messagesPrefixedWithArchitecture();
 	Treatment					picTreatment();
 	WeakReferenceMismatchTreatment	weakReferenceMismatchTreatment() const { return fWeakReferenceMismatchTreatment; }
@@ -266,7 +311,7 @@ public:
 	CommonsMode					commonsMode() const { return fCommonsMode; }
 	bool						warnCommons() const { return fWarnCommons; }
 	bool						keepRelocations();
-	FileInfo					findFile(const char* path) const;
+	FileInfo					findFile(const std::string &path) const;
 	UUIDMode					UUIDMode() const { return fUUIDMode; }
 	bool						warnStabs();
 	bool						pauseAtEnd() { return fPause; }
@@ -296,7 +341,7 @@ public:
 	const std::vector<DylibOverride>&	dylibOverrides() const { return fDylibOverrides; }
 	const char*					generatedMapPath() const { return fMapPath; }
 	bool						positionIndependentExecutable() const { return fPositionIndependentExecutable; }
-	Options::FileInfo			findFileUsingPaths(const char* path) const;
+	Options::FileInfo			findFileUsingPaths(const std::string &path) const;
 	bool						deadStripDylibs() const { return fDeadStripDylibs; }
 	bool						allowedUndefined(const char* name) const { return ( fAllowedUndefined.find(name) != fAllowedUndefined.end() ); }
 	bool						someAllowedUndefines() const { return (fAllowedUndefined.size() != 0); }
@@ -311,6 +356,7 @@ public:
 	bool						needsUnwindInfoSection() const { return fAddCompactUnwindEncoding; }
 	const std::vector<const char*>&	llvmOptions() const{ return fLLVMOptions; }
 	const std::vector<const char*>&	segmentOrder() const{ return fSegmentOrder; }
+	bool						segmentOrderAfterFixedAddressSegment(const char* segName) const;
 	const std::vector<const char*>* sectionOrder(const char* segName) const;
 	const std::vector<const char*>&	dyldEnvironExtras() const{ return fDyldEnvironExtras; }
 	const std::vector<const char*>&	astFilePaths() const{ return fASTFilePaths; }
@@ -346,7 +392,7 @@ public:
 	bool						objcGc() const { return fObjCGc; }
 	bool						objcGcOnly() const { return fObjCGcOnly; }
 	bool						canUseThreadLocalVariables() const { return fTLVSupport; }
-	bool						addVersionLoadCommand() const { return fVersionLoadCommand; }
+	bool						addVersionLoadCommand() const { return fVersionLoadCommand && (fPlatform != kPlatformUnknown); }
 	bool						addFunctionStarts() const { return fFunctionStartsLoadCommand; }
 	bool						addDataInCodeInfo() const { return fDataInCodeInfoLoadCommand; }
 	bool						canReExportSymbols() const { return fCanReExportSymbols; }
@@ -364,6 +410,15 @@ public:
 	bool						markAppExtensionSafe() const { return fMarkAppExtensionSafe; }
 	bool						checkDylibsAreAppExtensionSafe() const { return fCheckAppExtensionSafe; }
 	bool						forceLoadSwiftLibs() const { return fForceLoadSwiftLibs; }
+	bool						bundleBitcode() const { return fBundleBitcode; }
+	bool						hideSymbols() const { return fHideSymbols; }
+	bool						renameReverseSymbolMap() const { return fReverseMapUUIDRename; }
+	const char*					reverseSymbolMapPath() const { return fReverseMapPath; }
+	std::string					reverseMapTempPath() const { return fReverseMapTempPath; }
+	bool						ltoCodegenOnly() const { return fLTOCodegenOnly; }
+	bool						ignoreAutoLink() const { return fIgnoreAutoLink; }
+	bool						sharedRegionEncodingV2() const { return fSharedRegionEncodingV2; }
+	bool						useDataConstSegment() const { return fUseDataConstSegment; }
 	bool						hasWeakBitTweaks() const;
 	bool						forceWeak(const char* symbolName) const;
 	bool						forceNotWeak(const char* symbolName) const;
@@ -375,7 +430,6 @@ public:
 	bool						needsThreadLoadCommand() const { return fNeedsThreadLoadCommand; }
 	bool						needsEntryPointLoadCommand() const { return fEntryPointLoadCommand; }
 	bool						needsSourceVersionLoadCommand() const { return fSourceVersionLoadCommand; }
-	bool						needsDependentDRInfo() const { return fDependentDRInfo; }
 	bool						canUseAbsoluteSymbols() const { return fAbsoluteSymbols; }
 	bool						allowSimulatorToLinkWithMacOSX() const { return fAllowSimulatorToLinkWithMacOSX; }
 	uint64_t					sourceVersion() const { return fSourceVersion; }
@@ -390,10 +444,16 @@ public:
 								linkerOptions() const { return fLinkerOptions; }
 	FileInfo					findFramework(const char* frameworkName) const;
 	FileInfo					findLibrary(const char* rootName, bool dylibsOnly=false) const;
+	bool						armUsesZeroCostExceptions() const;
 	const std::vector<SectionRename>& sectionRenames() const { return fSectionRenames; }
 	const std::vector<SegmentRename>& segmentRenames() const { return fSegmentRenames; }
 	bool						moveRoSymbol(const char* symName, const char* filePath, const char*& seg, bool& wildCardMatch) const;
 	bool						moveRwSymbol(const char* symName, const char* filePath, const char*& seg, bool& wildCardMatch) const;
+	Platform					platform() const { return fPlatform; }
+	const std::vector<const char*>&	sdkPaths() const { return fSDKPaths; }
+	std::vector<std::string>	writeBitcodeLinkOptions() const;
+	std::string					getSDKVersionStr() const;
+	std::string					getPlatformStr() const;
 
 private:
 	typedef std::unordered_map<const char*, unsigned int, ld::CStringHash, ld::CStringEquals> NameToOrder;
@@ -413,6 +473,7 @@ private:
 		NameSet::iterator		regularBegin() const	{ return fRegular.begin(); }
 		NameSet::iterator		regularEnd() const		{ return fRegular.end(); }
 		void					remove(const NameSet&); 
+		std::vector<const char*>		data() const;
 	private:
 		static bool				hasWildCards(const char*);
 		bool					wildCardMatch(const char* pattern, const char* candidate) const;
@@ -436,6 +497,8 @@ private:
 											 FileInfo& result) const;
 	uint64_t					parseVersionNumber64(const char*);
 	uint32_t					parseVersionNumber32(const char*);
+	std::string					getVersionString32(uint32_t ver) const;
+	std::string					getVersionString64(uint64_t ver) const;
 	void						parseSectionOrderFile(const char* segment, const char* section, const char* path);
 	void						parseOrderFile(const char* path, bool cstring);
 	void						addSection(const char* segment, const char* section, const char* path);
@@ -449,6 +512,7 @@ private:
 	void						setUndefinedTreatment(const char* treatment);
 	void						setMacOSXVersionMin(const char* version);
 	void						setIOSVersionMin(const char* version);
+	void						setWatchOSVersionMin(const char* version);
 	void						setWeakReferenceMismatchTreatment(const char* treatment);
 	void						addDylibOverride(const char* paths);
 	void						addSectionAlignment(const char* segment, const char* section, const char* alignment);
@@ -464,6 +528,7 @@ private:
 	void						addSectionRename(const char* srcSegment, const char* srcSection, const char* dstSegment, const char* dstSection);
 	void						addSegmentRename(const char* srcSegment, const char* dstSegment);
 	void						addSymbolMove(const char* dstSegment, const char* symbolList, std::vector<SymbolsMove>& list, const char* optionName);
+	void						cannotBeUsedWithBitcode(const char* arg);
 
 
 //	ObjectFile::ReaderOptions			fReaderOptions;
@@ -549,6 +614,7 @@ private:
 	bool								fStatistics;
 	bool								fPrintOptions;
 	bool								fSharedRegionEligible;
+	bool								fSharedRegionEligibleForceOff;
 	bool								fPrintOrderFileStatistics;
 	bool								fReadOnlyx86Stubs;
 	bool								fPositionIndependentExecutable;
@@ -561,6 +627,8 @@ private:
 	bool								fKextsUseStubs;
 	bool								fUsingLazyDylibLinking;
 	bool								fEncryptable;
+	bool								fEncryptableForceOn;
+	bool								fEncryptableForceOff;
 	bool								fOrderData;
 	bool								fMarkDeadStrippableDylib;
 	bool								fMakeCompressedDyldInfo;
@@ -620,9 +688,6 @@ private:
 	bool								fSourceVersionLoadCommand;
 	bool								fSourceVersionLoadCommandForceOn;
 	bool								fSourceVersionLoadCommandForceOff;	
-	bool								fDependentDRInfo;
-	bool								fDependentDRInfoForcedOn;
-	bool								fDependentDRInfoForcedOff;
 	bool								fTargetIOSSimulator;
 	bool								fExportDynamic;
 	bool								fAbsoluteSymbols;
@@ -638,14 +703,27 @@ private:
 	bool								fMarkAppExtensionSafe;
 	bool								fCheckAppExtensionSafe;
 	bool								fForceLoadSwiftLibs;
+	bool								fSharedRegionEncodingV2;
+	bool								fUseDataConstSegment;
+	bool								fUseDataConstSegmentForceOn;
+	bool								fUseDataConstSegmentForceOff;
+	bool								fBundleBitcode;
+	bool								fHideSymbols;
+	bool								fReverseMapUUIDRename;
+	const char*							fReverseMapPath;
+	std::string							fReverseMapTempPath;
+	bool								fLTOCodegenOnly;
+	bool								fIgnoreAutoLink;
+	Platform							fPlatform;
 	DebugInfoStripping					fDebugInfoStripping;
 	const char*							fTraceOutputFile;
 	ld::MacVersionMin					fMacVersionMin;
 	ld::IOSVersionMin					fIOSVersionMin;
+	ld::WatchOSVersionMin				fWatchOSVersionMin;
 	std::vector<AliasPair>				fAliases;
 	std::vector<const char*>			fInitialUndefines;
 	NameSet								fAllowedUndefined;
-	NameSet								fWhyLive;
+	SetWithWildcards					fWhyLive;
 	std::vector<ExtraSection>			fExtraSections;
 	std::vector<SectionAlignment>		fSectionAlignments;
 	std::vector<OrderedSymbol>			fOrderedSymbols;
