@@ -33,6 +33,7 @@
 #include <pthread.h>
 #include <mach-o/dyld.h>
 #include <vector>
+#include <map>
 #include <unordered_set>
 #include <unordered_map>
 
@@ -118,6 +119,9 @@ public:
 																						_debugInfoModTime = modTime; 
 																						_cpuSubType = subtype;}
 
+    static bool                                         sSupportsLocalContext;
+    static bool                                         sHasTriedLocalContext;
+    bool                                                mergeIntoGenerator(lto_code_gen_t generator, bool useSetModule);
 private:
 	friend class Atom;
 	friend class InternalAtom;
@@ -128,6 +132,9 @@ private:
 	class Atom*								_atomArray;
 	uint32_t								_atomArrayCount;
 	lto_module_t							_module;
+	const char*                             _path;
+	const uint8_t*                          _content;
+	uint32_t                                _contentLength;
 	const char*								_debugInfoPath;
 	time_t									_debugInfoModTime;
 	ld::Section								_section;
@@ -302,8 +309,12 @@ ld::relocatable::File* Parser::parseMachOFile(const uint8_t* p, size_t len, cons
 	objOpts.forceDwarfConversion = false;
 	objOpts.neverConvertDwarf   = false;
 	objOpts.verboseOptimizationHints = options.verboseOptimizationHints;
-
+	objOpts.armUsesZeroCostExceptions = options.armUsesZeroCostExceptions;
+	objOpts.simulator			= options.simulator;
+	objOpts.ignoreMismatchPlatform = options.ignoreMismatchPlatform;
+	objOpts.platform			= options.platform;
 	objOpts.subType				= 0;
+	objOpts.srcKind				= ld::relocatable::File::kSourceLTO;
 	
 	// mach-o parsing is done in-memory, but need path for debug notes
 	const char* path = "/tmp/lto.o";
@@ -325,7 +336,8 @@ ld::relocatable::File* Parser::parseMachOFile(const uint8_t* p, size_t len, cons
 
 File::File(const char* pth, time_t mTime, ld::File::Ordinal ordinal, const uint8_t* content, uint32_t contentLength, cpu_type_t arch) 
 	: ld::relocatable::File(pth,mTime,ordinal), _architecture(arch), _internalAtom(*this), 
-	_atomArray(NULL), _atomArrayCount(0), _module(NULL), _debugInfoPath(pth), 
+	_atomArray(NULL), _atomArrayCount(0), _module(NULL), _path(pth),
+	_content(content), _contentLength(contentLength), _debugInfoPath(pth),
 	_section("__TEXT_", "__tmp_lto", ld::Section::typeTempLTO),
 	_fixupToInternal(0, ld::Fixup::k1of1, ld::Fixup::kindNone, &_internalAtom),
 	_debugInfo(ld::relocatable::File::kDebugInfoNone), _cpuSubType(0)
@@ -333,9 +345,19 @@ File::File(const char* pth, time_t mTime, ld::File::Ordinal ordinal, const uint8
 	const bool log = false;
 	
 	// create llvm module
+#if LTO_API_VERSION >= 11
+	if ( sSupportsLocalContext || !sHasTriedLocalContext ) {
+		_module = ::lto_module_create_in_local_context(content, contentLength, pth);
+	}
+	if ( !sHasTriedLocalContext ) {
+		sHasTriedLocalContext = true;
+		sSupportsLocalContext = (_module != NULL);
+	}
+	if ( (_module == NULL) && !sSupportsLocalContext )
+#endif
 #if LTO_API_VERSION >= 9
 	_module = ::lto_module_create_from_memory_with_path(content, contentLength, pth);
-	if ( _module == NULL )
+	if ( _module == NULL && !sSupportsLocalContext )
 #endif
 	_module = ::lto_module_create_from_memory(content, contentLength);
     if ( _module == NULL )
@@ -416,6 +438,11 @@ File::File(const char* pth, time_t mTime, ld::File::Ordinal ordinal, const uint8
 			if ( log ) fprintf(stderr, "\t%s (undefined)\n", name);
 		}
 	}
+
+#if LTO_API_VERSION >= 11
+	if ( sSupportsLocalContext )
+		this->release();
+#endif
 }
 
 File::~File()
@@ -423,6 +450,34 @@ File::~File()
 	this->release();
 }
 
+bool File::mergeIntoGenerator(lto_code_gen_t generator, bool useSetModule) {
+#if LTO_API_VERSION >= 11
+    if ( sSupportsLocalContext ) {
+        assert(!_module && "Expected module to be disposed");
+        _module = ::lto_module_create_in_codegen_context(_content, _contentLength,
+                                                        _path, generator);
+        if ( _module == NULL )
+            throwf("could not reparse object file %s: '%s', using libLTO version '%s'",
+                  _path, ::lto_get_error_message(), ::lto_get_version());
+    }
+#endif
+    assert(_module && "Expected module to stick around");
+#if LTO_API_VERSION >= 13
+    if (useSetModule) {
+        // lto_codegen_set_module will transfer ownership of the module to LTO code generator,
+        // so we don't need to release the module here.
+        ::lto_codegen_set_module(generator, _module);
+        return false;
+    }
+#endif
+    if ( ::lto_codegen_add_module(generator, _module) )
+        return true;
+
+    // <rdar://problem/15471128> linker should release module as soon as possible
+    this->release();
+    return false;
+}
+
 void File::release()
 {
 	if ( _module != NULL )
@@ -519,7 +574,13 @@ bool Parser::optimize(  const std::vector<const ld::Atom*>&	allAtoms,
 		fprintf(stderr, "%s\n", ::lto_get_version());
 	
 	// create optimizer and add each Reader
-	lto_code_gen_t generator = ::lto_codegen_create();
+	lto_code_gen_t generator = NULL;
+#if LTO_API_VERSION >= 11
+	if ( File::sSupportsLocalContext )
+		generator = ::lto_codegen_create_in_local_context();
+	else
+#endif
+		generator = ::lto_codegen_create();
 #if LTO_API_VERSION >= 7
 	lto_codegen_set_diagnostic_handler(generator, ltoDiagnosticHandler, NULL);
 #endif
@@ -527,14 +588,20 @@ bool Parser::optimize(  const std::vector<const ld::Atom*>&	allAtoms,
 	// <rdar://problem/12379604> The order that files are merged must match command line order
 	std::sort(_s_files.begin(), _s_files.end(), CommandLineOrderFileSorter());
 	ld::File::Ordinal lastOrdinal;
+
+	// When flto_codegen_only is on and we have a single .bc file, use lto_codegen_set_module instead of
+	// lto_codegen_add_module, to make sure the the destination module will be the same as the input .bc file.
+	bool useSetModule = false;
+#if LTO_API_VERSION >= 13
+	useSetModule = (_s_files.size() == 1) && options.ltoCodegenOnly && (::lto_api_version() >= 13);
+#endif
 	for (std::vector<File*>::iterator it=_s_files.begin(); it != _s_files.end(); ++it) {
 		File* f = *it;
 		assert(f->ordinal() > lastOrdinal);
-		if ( logBitcodeFiles ) fprintf(stderr, "lto_codegen_add_module(%s)\n", f->path());
-		if ( ::lto_codegen_add_module(generator, f->module()) )
+		if ( logBitcodeFiles && !useSetModule) fprintf(stderr, "lto_codegen_add_module(%s)\n", f->path());
+		if ( logBitcodeFiles && useSetModule) fprintf(stderr, "lto_codegen_set_module(%s)\n", f->path());
+		if ( f->mergeIntoGenerator(generator, useSetModule) )
 			throwf("lto: could not merge in %s because '%s', using libLTO version '%s'", f->path(), ::lto_get_error_message(), ::lto_get_version());
-		// <rdar://problem/15471128> linker should release module as soon as possible
-		f->release();
 		lastOrdinal = f->ordinal();
 	}
 
@@ -654,6 +721,9 @@ bool Parser::optimize(  const std::vector<const ld::Atom*>&	allAtoms,
 
     // special case running ld -r on all bitcode files to produce another bitcode file (instead of mach-o)
     if ( options.relocatable && !hasNonllvmAtoms ) {
+#if LTO_API_VERSION >= 15
+		::lto_codegen_set_should_embed_uselists(generator, false);
+#endif
 		if ( ! ::lto_codegen_write_merged_modules(generator, options.outputFilePath) ) {
 			// HACK, no good way to tell linker we are all done, so just quit
 			exit(0);
@@ -692,6 +762,9 @@ bool Parser::optimize(  const std::vector<const ld::Atom*>&	allAtoms,
         char tempBitcodePath[MAXPATHLEN];
         strcpy(tempBitcodePath, options.outputFilePath);
         strcat(tempBitcodePath, ".lto.bc");
+#if LTO_API_VERSION >= 15
+        ::lto_codegen_set_should_embed_uselists(generator, true);
+#endif
         ::lto_codegen_write_merged_modules(generator, tempBitcodePath);
     }
 
@@ -709,11 +782,62 @@ bool Parser::optimize(  const std::vector<const ld::Atom*>&	allAtoms,
 		}
 	}
 #endif
-	// run code generator
-	size_t machOFileLen;
-	const uint8_t* machOFile = (uint8_t*)::lto_codegen_compile(generator, &machOFileLen);
-	if ( machOFile == NULL ) 
-		throwf("could not do LTO codegen: '%s', using libLTO version '%s'", ::lto_get_error_message(), ::lto_get_version());
+
+	// When lto API version is greater than or equal to 12, we use lto_codegen_optimize and lto_codegen_compile_optimized
+	// instead of lto_codegen_compile, and we save the merged bitcode file in between.
+	bool useSplitAPI = false;
+#if LTO_API_VERSION >= 12
+	if ( ::lto_api_version() >= 12)
+		useSplitAPI = true;
+#endif
+
+	size_t machOFileLen = 0;
+	const uint8_t* machOFile = NULL;
+	if ( useSplitAPI) {
+#if LTO_API_VERSION >= 12
+#if LTO_API_VERSION >= 14
+	if ( ::lto_api_version() >= 14 && options.ltoCodegenOnly)
+          lto_codegen_set_should_internalize(generator, false);
+#endif
+		// run optimizer
+		if ( !options.ltoCodegenOnly && ::lto_codegen_optimize(generator) )
+			throwf("could not do LTO optimization: '%s', using libLTO version '%s'", ::lto_get_error_message(), ::lto_get_version());
+
+		if ( options.saveTemps || options.bitcodeBundle ) {
+			// save off merged bitcode file
+			char tempOptBitcodePath[MAXPATHLEN];
+			strcpy(tempOptBitcodePath, options.outputFilePath);
+			strcat(tempOptBitcodePath, ".lto.opt.bc");
+#if LTO_API_VERSION >= 15
+			::lto_codegen_set_should_embed_uselists(generator, true);
+#endif
+			::lto_codegen_write_merged_modules(generator, tempOptBitcodePath);
+			if ( options.bitcodeBundle )
+				state.ltoBitcodePath = tempOptBitcodePath;
+		}
+
+		// run code generator
+		machOFile = (uint8_t*)::lto_codegen_compile_optimized(generator, &machOFileLen);
+#endif
+		if ( machOFile == NULL )
+			throwf("could not do LTO codegen: '%s', using libLTO version '%s'", ::lto_get_error_message(), ::lto_get_version());
+	}
+	else {
+		// run optimizer and code generator
+		machOFile = (uint8_t*)::lto_codegen_compile(generator, &machOFileLen);
+		if ( machOFile == NULL )
+			throwf("could not do LTO codegen: '%s', using libLTO version '%s'", ::lto_get_error_message(), ::lto_get_version());
+		if ( options.saveTemps ) {
+			// save off merged bitcode file
+			char tempOptBitcodePath[MAXPATHLEN];
+			strcpy(tempOptBitcodePath, options.outputFilePath);
+			strcat(tempOptBitcodePath, ".lto.opt.bc");
+#if LTO_API_VERSION >= 15
+			::lto_codegen_set_should_embed_uselists(generator, true);
+#endif
+			::lto_codegen_write_merged_modules(generator, tempOptBitcodePath);
+		}
+	}
 	
     // if requested, save off temp mach-o file
     if ( options.saveTemps ) {
@@ -725,11 +849,6 @@ bool Parser::optimize(  const std::vector<const ld::Atom*>&	allAtoms,
 			::write(fd, machOFile, machOFileLen);
 			::close(fd);
 		}
-		//	save off merged bitcode file
-		char tempOptBitcodePath[MAXPATHLEN];
-        strcpy(tempOptBitcodePath, options.outputFilePath);
-        strcat(tempOptBitcodePath, ".lto.opt.bc");
-        ::lto_codegen_write_merged_modules(generator, tempOptBitcodePath);
 	}
 
 	// if needed, save temp mach-o file to specific location
@@ -861,7 +980,7 @@ void Parser::AtomSyncer::doAtom(const ld::Atom& machoAtom)
 					}
 					else {
 						// <rdar://problem/12859831> Don't unbind follow-on reference into by-name reference 
-						if ( (_deadllvmAtoms.find(targetName) != _deadllvmAtoms.end()) && (fit->kind != ld::Fixup::kindNoneFollowOn) ) {
+						if ( (_deadllvmAtoms.find(targetName) != _deadllvmAtoms.end()) && (fit->kind != ld::Fixup::kindNoneFollowOn) && (fit->u.target->scope() != ld::Atom::scopeTranslationUnit) ) {
 							// target was coalesed away and replace by mach-o atom from a non llvm .o file
 							fit->binding = ld::Fixup::bindingByNameUnbound;
 							fit->u.name = targetName;
@@ -889,6 +1008,8 @@ public:
 	~Mutex() { pthread_mutex_unlock(&lto_lock); }
 };
 pthread_mutex_t Mutex::lto_lock = PTHREAD_MUTEX_INITIALIZER;
+bool File::sSupportsLocalContext = false;
+bool File::sHasTriedLocalContext = false;
 
 //
 // Used by archive reader to see if member is an llvm bitcode file
@@ -900,6 +1021,18 @@ bool isObjectFile(const uint8_t* fileContent, uint64_t fileLength, cpu_type_t ar
 }
 
 
+static ld::relocatable::File *parseImpl(
+          const uint8_t *fileContent, uint64_t fileLength, const char *path,
+          time_t modTime, ld::File::Ordinal ordinal, cpu_type_t architecture,
+          cpu_subtype_t subarch, bool logAllFiles,
+          bool verboseOptimizationHints)
+{
+	if ( Parser::validFile(fileContent, fileLength, architecture, subarch) )
+		return Parser::parse(fileContent, fileLength, path, modTime, ordinal, architecture, subarch, logAllFiles, verboseOptimizationHints);
+	else
+		return NULL;
+}
+
 //
 // main function used by linker to instantiate ld::Files
 //
@@ -908,11 +1041,12 @@ ld::relocatable::File* parse(const uint8_t* fileContent, uint64_t fileLength,
 								cpu_type_t architecture, cpu_subtype_t subarch, bool logAllFiles,
 								bool verboseOptimizationHints)
 {
+	// Note: Once lto_module_create_in_local_context() and friends are thread safe
+	// this lock can be removed.
 	Mutex lock;
-	if ( Parser::validFile(fileContent, fileLength, architecture, subarch) )
-		return Parser::parse(fileContent, fileLength, path, modTime, ordinal, architecture, subarch, logAllFiles, verboseOptimizationHints);
-	else
-		return NULL;
+	return parseImpl(fileContent, fileLength, path, modTime, ordinal,
+					architecture, subarch, logAllFiles,
+					verboseOptimizationHints);
 }
 
 //
