@@ -32,15 +32,17 @@
 
 #include <vector>
 #include <set>
+#include <map>
 #include <algorithm>
-#include <ext/hash_map>
-#include <ext/hash_set>
+#include <unordered_map>
+#include <unordered_set>
 
 #include "Architectures.hpp"
+#include "Bitcode.hpp"
 #include "MachOFileAbstraction.hpp"
 #include "MachOTrie.hpp"
 #include "macho_dylib_file.h"
-
+#include "../code-sign-blobs/superblob.h"
 
 namespace mach_o {
 namespace dylib {
@@ -68,8 +70,6 @@ public:
 													_file(f), _name(nm), _address(address) {}
 	// overrides of ld::Atom
 	virtual const ld::File*						file() const		{ return &_file; }
-	virtual bool								translationUnitSource(const char** dir, const char** nm) const
-																	{ return false; }
 	virtual const char*							name() const		{ return _name; }
 	virtual uint64_t							size() const		{ return 0; }
 	virtual uint64_t							objectAddress() const { return _address; }
@@ -101,8 +101,6 @@ public:
 
 	// overrides of ld::Atom
 	virtual ld::File*							file() const		{ return &_file; }
-	virtual bool								translationUnitSource(const char** dir, const char** nm) const
-																	{ return false; }
 	virtual const char*							name() const		{ return "import-atom"; }
 	virtual uint64_t							size() const		{ return 0; }
 	virtual uint64_t							objectAddress() const { return 0; }
@@ -144,17 +142,22 @@ class File : public ld::dylib::File
 public:
 	static bool								validFile(const uint8_t* fileContent, bool executableOrDylib);
 											File(const uint8_t* fileContent, uint64_t fileLength, const char* path,   
-													time_t mTime, uint32_t ordinal, bool linkingFlatNamespace, 
+													time_t mTime, ld::File::Ordinal ordinal, bool linkingFlatNamespace, 
 													bool linkingMainExecutable, bool hoistImplicitPublicDylibs, 
-													ld::MacVersionMin macMin, ld::IOSVersionMin iPhoneMin, bool addVers, 
-													bool logAllFiles, const char* installPath, bool indirectDylib);
+													Options::Platform platform, uint32_t linkMinOSVersion, bool allowSimToMacOSX,
+													bool addVers,  bool buildingForSimulator,
+													bool logAllFiles, const char* installPath,
+													bool indirectDylib, bool ignoreMismatchPlatform);
 	virtual									~File() {}
 
 	// overrides of ld::File
 	virtual bool							forEachAtom(ld::File::AtomHandler&) const;
 	virtual bool							justInTimeforEachAtom(const char* name, ld::File::AtomHandler&) const;
 	virtual ld::File::ObjcConstraint		objCConstraint() const		{ return _objcContraint; }
-	
+	virtual uint8_t							swiftVersion() const		{ return _swiftVersion; }
+	virtual uint32_t						minOSVersion() const		{ return _minVersionInDylib; }
+	virtual uint32_t						platformLoadCommand() const	{ return _platformInDylib; }
+
 	// overrides of ld::dylib::File
 	virtual void							processIndirectLibraries(ld::dylib::File::DylibHandler*, bool);
 	virtual bool							providedExportAtom() const	{ return _providedAtom; }
@@ -165,13 +168,12 @@ public:
 	virtual bool							hasPublicInstallName() const{ return _hasPublicInstallName; }
 	virtual bool							hasWeakDefinition(const char* name) const;
 	virtual bool							allSymbolsAreWeakImported() const;
-
+	virtual bool							installPathVersionSpecific() const { return _installPathOverride; }
+	virtual bool							appExtensionSafe() const	{ return _appExtensionSafe; };
+	virtual ld::Bitcode*					getBitcode() const			{ return _bitcode.get(); }
 
 protected:
-
-	struct ReExportChain { ReExportChain* prev; File<A>* file; };
-
-	void											assertNoReExportCycles(ReExportChain*);
+	virtual void							assertNoReExportCycles(ReExportChain*) const;
 
 private:
 	typedef typename A::P					P;
@@ -181,19 +183,24 @@ private:
 	friend class ExportAtom<A>;
 	friend class ImportAtom<A>;
 
-	class CStringEquals
-	{
-	public:
-		bool operator()(const char* left, const char* right) const { return (strcmp(left, right) == 0); }
+	struct CStringHash {
+		std::size_t operator()(const char* __s) const {
+			unsigned long __h = 0;
+			for ( ; *__s; ++__s)
+				__h = 5 * __h + *__s;
+			return size_t(__h);
+		};
 	};
-	struct AtomAndWeak { ld::Atom* atom; bool weakDef; bool tlv; pint_t address; };
-	typedef __gnu_cxx::hash_map<const char*, AtomAndWeak, __gnu_cxx::hash<const char*>, CStringEquals> NameToAtomMap;
-	typedef __gnu_cxx::hash_set<const char*, __gnu_cxx::hash<const char*>, CStringEquals>  NameSet;
+	struct AtomAndWeak { ld::Atom* atom; bool weakDef; bool tlv; uint64_t address; };
+	typedef std::unordered_map<const char*, AtomAndWeak, ld::CStringHash, ld::CStringEquals> NameToAtomMap;
+	typedef std::unordered_set<const char*, CStringHash, ld::CStringEquals>  NameSet;
 
 	struct Dependent { const char* path; File<A>* dylib; bool reExport; };
 
-	bool										containsOrReExports(const char* name, bool* weakDef, bool* tlv, pint_t* defAddress) const;
+	virtual std::pair<bool, bool>				hasWeakDefinitionImpl(const char* name) const;
+	virtual bool								containsOrReExports(const char* name, bool& weakDef, bool& tlv, uint64_t& defAddress) const;
 	bool										isPublicLocation(const char* pth);
+	bool										wrongOS() { return _wrongOS; }
 	void										addSymbol(const char* name, bool weak, bool tlv, pint_t address);
 	void										addDyldFastStub();
 	void										buildExportHashTableFromExportInfo(const macho_dyld_info_command<P>* dyldInfo,
@@ -201,15 +208,18 @@ private:
 	void										buildExportHashTableFromSymbolTable(const macho_dysymtab_command<P>* dynamicInfo, 
 														const macho_nlist<P>* symbolTable, const char* strings,
 														const uint8_t* fileContent);
+	static uint32_t								parseVersionNumber32(const char* versionString);
 	static const char*							objCInfoSegmentName();
 	static const char*							objCInfoSectionName();
-	
-	const ld::MacVersionMin						_macVersionMin;
-	const ld::IOSVersionMin						_iOSVersionMin;
+
+	const Options::Platform						_platform;
+	const uint32_t								_linkMinOSVersion;
+	const bool									_allowSimToMacOSXLinking;
 	const bool									_addVersionLoadCommand;
 	bool										_linkingFlat;
 	bool										_implicitlyLinkPublicDylibs;
 	ld::File::ObjcConstraint					_objcContraint;
+	uint8_t										_swiftVersion;
 	ld::Section									_importProxySection;
 	ld::Section									_flatDummySection;
 	std::vector<Dependent>						_dependentDylibs;
@@ -224,7 +234,14 @@ private:
 	bool										_hasPublicInstallName;
 	mutable bool								_providedAtom;
 	bool										_explictReExportFound;
-
+	bool										_wrongOS;
+	bool										_installPathOverride;
+	bool										_indirectDylibsProcessed;
+	bool										_appExtensionSafe;
+	uint32_t									_minVersionInDylib;
+	uint32_t									_platformInDylib;
+	std::unique_ptr<ld::Bitcode>				_bitcode;
+	
 	static bool									_s_logHashtable;
 };
 
@@ -240,19 +257,22 @@ template <> const char* File<arm>::objCInfoSectionName() { return "__objc_imagei
 template <typename A> const char* File<A>::objCInfoSectionName() { return "__image_info"; }
 
 template <typename A>
-File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth, time_t mTime, uint32_t ord,
+File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth, time_t mTime, ld::File::Ordinal ord,
 				bool linkingFlatNamespace, bool linkingMainExecutable, bool hoistImplicitPublicDylibs,
-				ld::MacVersionMin macMin, ld::IOSVersionMin iOSMin, bool addVers,
-				bool logAllFiles, const char* targetInstallPath, bool indirectDylib)
+				Options::Platform platform, uint32_t linkMinOSVersion, bool allowSimToMacOSX, bool addVers, bool buildingForSimulator,
+				bool logAllFiles, const char* targetInstallPath, bool indirectDylib, bool ignoreMismatchPlatform)
 	: ld::dylib::File(strdup(pth), mTime, ord), 
-	_macVersionMin(macMin), _iOSVersionMin(iOSMin), _addVersionLoadCommand(addVers), 
+	_platform(platform), _linkMinOSVersion(linkMinOSVersion), _allowSimToMacOSXLinking(allowSimToMacOSX), _addVersionLoadCommand(addVers), 
 	_linkingFlat(linkingFlatNamespace), _implicitlyLinkPublicDylibs(hoistImplicitPublicDylibs),
-	_objcContraint(ld::File::objcConstraintNone),
+	_objcContraint(ld::File::objcConstraintNone), _swiftVersion(0),
 	_importProxySection("__TEXT", "__import", ld::Section::typeImportProxies, true),
 	_flatDummySection("__LINKEDIT", "__flat_dummy", ld::Section::typeLinkEdit, true),
-	_parentUmbrella(NULL), _importAtom(NULL), _noRexports(false), _hasWeakExports(false), 
+	_parentUmbrella(NULL), _importAtom(NULL),
+	_noRexports(false), _hasWeakExports(false), 
 	_deadStrippable(false), _hasPublicInstallName(false), 
-	 _providedAtom(false), _explictReExportFound(false)
+	 _providedAtom(false), _explictReExportFound(false), _wrongOS(false), _installPathOverride(false), 
+	_indirectDylibsProcessed(false), _appExtensionSafe(false),
+	_minVersionInDylib(0), _platformInDylib(Options::kPlatformUnknown)
 {
 	const macho_header<P>* header = (const macho_header<P>*)fileContent;
 	const uint32_t cmd_count = header->ncmds();
@@ -277,6 +297,7 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 					|| (header->filetype() == MH_EXECUTE);  // bundles and exectuables can be used via -bundle_loader
 	_hasWeakExports = (header->flags() & MH_WEAK_DEFINES);
 	_deadStrippable = (header->flags() & MH_DEAD_STRIPPABLE_DYLIB);
+	_appExtensionSafe = (header->flags() & MH_APP_EXTENSION_SAFE);
 	
 	// pass 1: get pointers, and see if this dylib uses compressed LINKEDIT format
 	const macho_dysymtab_command<P>* dynamicInfo = NULL;
@@ -285,6 +306,7 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 	const char*	strings = NULL;
 	bool compressedLinkEdit = false;
 	uint32_t dependentLibCount = 0;
+	Options::Platform lcPlatform = Options::kPlatformUnknown;
 	const macho_load_command<P>* cmd = cmds;
 	for (uint32_t i = 0; i < cmd_count; ++i) {
 		macho_dylib_command<P>* dylibID;
@@ -294,6 +316,8 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 				symtab = (macho_symtab_command<P>*)cmd;
 				symbolTable = (const macho_nlist<P>*)((char*)header + symtab->symoff());
 				strings = (char*)header + symtab->stroff();
+				if ( (symtab->stroff() + symtab->strsize()) > fileLength )
+					throwf("mach-o string pool extends beyond end of file in %s", pth);
 				break;
 			case LC_DYSYMTAB:
 				dynamicInfo = (macho_dysymtab_command<P>*)cmd;
@@ -324,18 +348,24 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 				break;
 			case LC_SUB_CLIENT:
 				_allowableClients.push_back(strdup(((macho_sub_client_command<P>*)cmd)->client()));
+				// <rdar://problem/20627554> Don't hoist "public" (in /usr/lib/) dylibs that should not be directly linked
+				_hasPublicInstallName = false;
 				break;
 			case LC_VERSION_MIN_MACOSX:
-				if ( _addVersionLoadCommand && !indirectDylib && (_iOSVersionMin != ld::iOSVersionUnset) )
-					warning("building for iOS, but linking against dylib built for MacOSX: %s", pth);
-				break;
 			case LC_VERSION_MIN_IPHONEOS:
-				if ( _addVersionLoadCommand && !indirectDylib && (_macVersionMin != ld::macVersionUnset) )
-					warning("building for MacOSX, but linking against dylib built for iOS: %s", pth);
+			case LC_VERSION_MIN_WATCHOS:
+	#if SUPPORT_APPLE_TV
+			case LC_VERSION_MIN_TVOS:
+	#endif
+				_minVersionInDylib = (ld::MacVersionMin)((macho_version_min_command<P>*)cmd)->version();
+				_platformInDylib = cmd->cmd();
+				lcPlatform = Options::platformForLoadCommand(_platformInDylib);
+				break;
+			case LC_CODE_SIGNATURE:
 				break;
 			case macho_segment_command<P>::CMD:
 				// check for Objective-C info
-				if ( strcmp(((macho_segment_command<P>*)cmd)->segname(), objCInfoSegmentName()) == 0 ) {
+				if ( strncmp(((macho_segment_command<P>*)cmd)->segname(), objCInfoSegmentName(), 6) == 0 ) {
 					const macho_segment_command<P>* segment = (macho_segment_command<P>*)cmd;
 					const macho_section<P>* const sectionsStart = (macho_section<P>*)((char*)segment + sizeof(macho_segment_command<P>));
 					const macho_section<P>* const sectionsEnd = &sectionsStart[segment->nsects()];
@@ -347,6 +377,7 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 							//	};
 							// #define OBJC_IMAGE_SUPPORTS_GC   2
 							// #define OBJC_IMAGE_GC_ONLY       4
+						        // #define OBJC_IMAGE_IS_SIMULATED  32
 							//
 							const uint32_t* contents = (uint32_t*)(&fileContent[sect->offset()]);
 							if ( (sect->size() >= 8) && (contents[0] == 0) ) {
@@ -355,8 +386,11 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 									_objcContraint = ld::File::objcConstraintGC;
 								else if ( (flags & 2) == 2 )
 									_objcContraint = ld::File::objcConstraintRetainReleaseOrGC;
+								else if ( (flags & 32) == 32 )
+									_objcContraint = ld::File::objcConstraintRetainReleaseForSimulator;
 								else
 									_objcContraint = ld::File::objcConstraintRetainRelease;
+								_swiftVersion = ((flags >> 8) & 0xFF);
 							}
 							else if ( sect->size() > 0 ) {
 								warning("can't parse %s/%s section in %s", objCInfoSegmentName(), objCInfoSectionName(), this->path());
@@ -364,11 +398,88 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 						}
 					}
 				}
+				// Construct bitcode if there is a bitcode bundle section in the dylib
+				// Record the size of the section because the content is not checked
+				else if ( strcmp(((macho_segment_command<P>*)cmd)->segname(), "__LLVM") == 0 ) {
+					const macho_section<P>* const sect = (macho_section<P>*)((char*)cmd + sizeof(macho_segment_command<P>));
+					if ( strncmp(sect->sectname(), "__bundle", 8) == 0 )
+						_bitcode = std::unique_ptr<ld::Bitcode>(new ld::Bitcode(NULL, sect->size()));
+				}
 		}
 		cmd = (const macho_load_command<P>*)(((char*)cmd)+cmd->cmdsize());
 		if ( cmd > cmdsEnd )
 			throwf("malformed dylb, load command #%d is outside size of load commands in %s", i, pth);
 	}
+	// arm/arm64 objects are default to ios platform if not set.
+	// rdar://problem/21746314
+	if (lcPlatform == Options::kPlatformUnknown &&
+		(std::is_same<A, arm>::value || std::is_same<A, arm64>::value))
+		lcPlatform = Options::kPlatformiOS;
+
+	// check cross-linking
+	if ( lcPlatform != platform ) {
+		_wrongOS = true;
+		if ( _addVersionLoadCommand && !indirectDylib && !ignoreMismatchPlatform ) {
+			if ( buildingForSimulator ) {
+				if ( !_allowSimToMacOSXLinking ) {
+					switch (platform) {
+						case Options::kPlatformOSX:
+						case Options::kPlatformiOS:
+							if ( lcPlatform == Options::kPlatformUnknown )
+								break;
+							// fall through if the Platform is not Unknown
+						case Options::kPlatformWatchOS:
+							// WatchOS errors on cross-linking all the time.
+							throwf("building for %s simulator, but linking against dylib built for %s,",
+								   Options::platformName(platform),
+								   Options::platformName(lcPlatform));
+							break;
+	#if SUPPORT_APPLE_TV
+						case Options::kPlatform_tvOS:
+							// tvOS is a warning temporarily. rdar://problem/21746965
+							if (platform == Options::kPlatform_tvOS)
+								warning("URGENT: building for %s simulator, but linking against dylib (%s) built for %s. "
+										"Note: This will be an error in the future.",
+										Options::platformName(platform), path(),
+										Options::platformName(lcPlatform));
+							break;
+	#endif
+						case Options::kPlatformUnknown:
+							// skip if the target platform is unknown
+							break;
+					}
+				}
+			}
+			else {
+				switch (platform) {
+					case Options::kPlatformOSX:
+					case Options::kPlatformiOS:
+						if ( lcPlatform == Options::kPlatformUnknown )
+							break;
+						// fall through if the Platform is not Unknown
+					case Options::kPlatformWatchOS:
+						// WatchOS errors on cross-linking all the time.
+						throwf("building for %s, but linking against dylib built for %s,",
+							   Options::platformName(platform),
+							   Options::platformName(lcPlatform));
+						break;
+	#if SUPPORT_APPLE_TV
+					case Options::kPlatform_tvOS:
+						// tvOS is a warning temporarily. rdar://problem/21746965
+						if (platform == Options::kPlatform_tvOS)
+							warning("URGENT: building for %s, but linking against dylib (%s) built for %s. "
+									"Note: This will be an error in the future.",
+									Options::platformName(platform), path(),
+									Options::platformName(lcPlatform));
+						break;
+	#endif
+					case Options::kPlatformUnknown:
+						// skip if the target platform is unknown
+						break;
+				}
+			}
+		}
+	}
 
 	// figure out if we need to examine dependent dylibs
 	// with compressed LINKEDIT format, MH_NO_REEXPORTED_DYLIBS can be trusted
@@ -380,6 +491,7 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 		// pass 2 builds list of all dependent libraries
 		_dependentDylibs.reserve(dependentLibCount);
 		cmd = cmds;
+		unsigned int reExportDylibCount = 0;  
 		for (uint32_t i = 0; i < cmd_count; ++i) {
 			switch (cmd->cmd()) {
 				case LC_LOAD_DYLIB:
@@ -388,6 +500,7 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 					if ( compressedLinkEdit && !linkingFlatNamespace ) 
 						break;
 				case LC_REEXPORT_DYLIB:
+					++reExportDylibCount;
 					Dependent entry;
 					entry.path = strdup(((macho_dylib_command<P>*)cmd)->name());
 					entry.dylib = NULL;
@@ -400,7 +513,8 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 		}
 		// verify MH_NO_REEXPORTED_DYLIBS bit was correct
 		if ( compressedLinkEdit && !linkingFlatNamespace ) {
-			assert(_dependentDylibs.size() != 0);
+			if ( reExportDylibCount == 0 )
+				throwf("malformed dylib has MH_NO_REEXPORTED_DYLIBS flag but no LC_REEXPORT_DYLIB load commands: %s", pth);
 		}
 		// pass 3 add re-export info
 		cmd = cmds;
@@ -459,7 +573,7 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 		}
 		_importAtom = new ImportAtom<A>(*this, importNames);
 	}
-	
+
 	// build hash table
 	if ( dyldInfo != NULL ) 
 		buildExportHashTableFromExportInfo(dyldInfo, fileContent);
@@ -471,6 +585,29 @@ File<A>::File(const uint8_t* fileContent, uint64_t fileLength, const char* pth,
 }
 
 
+//
+// Parses number of form X[.Y[.Z]] into a uint32_t where the nibbles are xxxx.yy.zz
+//
+template <typename A>
+uint32_t File<A>::parseVersionNumber32(const char* versionString)
+{
+	uint32_t x = 0;
+	uint32_t y = 0;
+	uint32_t z = 0;
+	char* end;
+	x = strtoul(versionString, &end, 10);
+	if ( *end == '.' ) {
+		y = strtoul(&end[1], &end, 10);
+		if ( *end == '.' ) {
+			z = strtoul(&end[1], &end, 10);
+		}
+	}
+	if ( (*end != '\0') || (x > 0xffff) || (y > 0xff) || (z > 0xff) )
+		throwf("malformed 32-bit x.y.z version number: %s", versionString);
+
+	return (x << 16) | ( y << 8 ) | z;
+}
+
 template <typename A>
 void File<A>::buildExportHashTableFromSymbolTable(const macho_dysymtab_command<P>* dynamicInfo, 
 												const macho_nlist<P>* symbolTable, const char* strings, 
@@ -480,14 +617,14 @@ void File<A>::buildExportHashTableFromSymbolTable(const macho_dysymtab_command<P
 		if ( _s_logHashtable ) fprintf(stderr, "ld: building hashtable of %u toc entries for %s\n", dynamicInfo->nextdefsym(), this->path());
 		const macho_nlist<P>* start = &symbolTable[dynamicInfo->iextdefsym()];
 		const macho_nlist<P>* end = &start[dynamicInfo->nextdefsym()];
-		_atoms.resize(dynamicInfo->nextdefsym()); // set initial bucket count
+		_atoms.reserve(dynamicInfo->nextdefsym()); // set initial bucket count
 		for (const macho_nlist<P>* sym=start; sym < end; ++sym) {
 			this->addSymbol(&strings[sym->n_strx()], (sym->n_desc() & N_WEAK_DEF) != 0, false, sym->n_value());
 		}
 	}
 	else {
 		int32_t count = dynamicInfo->ntoc();
-		_atoms.resize(count); // set initial bucket count
+		_atoms.reserve(count); // set initial bucket count
 		if ( _s_logHashtable ) fprintf(stderr, "ld: building hashtable of %u entries for %s\n", count, this->path());
 		const struct dylib_table_of_contents* toc = (dylib_table_of_contents*)(fileContent + dynamicInfo->tocoff());
 		for (int32_t i = 0; i < count; ++i) {
@@ -543,9 +680,6 @@ void File<A>::addDyldFastStub()
 template <typename A>
 void File<A>::addSymbol(const char* name, bool weakDef, bool tlv, pint_t address)
 {
-	if ( weakDef ) {
-		assert(_hasWeakExports);
-	}
 	//fprintf(stderr, "addSymbol() %s\n", name);
 	// symbols that start with $ld$ are meta-data to the static linker
 	// <rdar://problem/5182537> need way for ld and dyld to see different exported symbols in a dylib
@@ -555,15 +689,7 @@ void File<A>::addSymbol(const char* name, bool weakDef, bool tlv, pint_t address
 		const char* symCond = strchr(symAction, '$');
 		if ( symCond != NULL ) {
 			char curOSVers[16];
-			if ( _macVersionMin != ld::macVersionUnset ) {
-				sprintf(curOSVers, "$os%d.%d$", (_macVersionMin >> 16), ((_macVersionMin >> 8) & 0xFF));
-			}
-			else if ( _iOSVersionMin != ld::iOSVersionUnset ) {
-				sprintf(curOSVers, "$os%d.%d$", (_iOSVersionMin >> 16), ((_iOSVersionMin >> 8) & 0xFF));
-			}
-			else {
-				assert(0 && "targeting neither macosx nor iphoneos");
-			}
+			sprintf(curOSVers, "$os%d.%d$", (_linkMinOSVersion >> 16), ((_linkMinOSVersion >> 8) & 0xFF));
 			if ( strncmp(symCond, curOSVers, strlen(curOSVers)) == 0 ) {
 				const char* symName = strchr(&symCond[1], '$');
 				if ( symName != NULL ) {
@@ -577,6 +703,18 @@ void File<A>::addSymbol(const char* name, bool weakDef, bool tlv, pint_t address
 						this->addSymbol(symName, weakDef, false, 0);
 						return;
 					}
+					else if ( strncmp(symAction, "install_name$", 13) == 0 ) {
+						_dylibInstallPath = symName;
+						_installPathOverride = true;
+						// <rdar://problem/14448206> CoreGraphics redirects to ApplicationServices, but with wrong compat version
+						if ( strcmp(_dylibInstallPath, "/System/Library/Frameworks/ApplicationServices.framework/Versions/A/ApplicationServices") == 0 )
+							_dylibCompatibilityVersion = parseVersionNumber32("1.0");
+						return;
+					}
+					else if ( strncmp(symAction, "compatibility_version$", 22) == 0 ) {
+						_dylibCompatibilityVersion = parseVersionNumber32(symName);
+						return;
+					}
 					else {
 						warning("bad symbol action: %s in dylib %s", name, this->path());
 					}
@@ -614,29 +752,34 @@ bool File<A>::forEachAtom(ld::File::AtomHandler& handler) const
 	return false;
 }
 
+
+template <typename A>
+std::pair<bool, bool> File<A>::hasWeakDefinitionImpl(const char* name) const
+{
+	const auto pos = _atoms.find(name);
+	if ( pos != _atoms.end() )
+		return std::make_pair(true, pos->second.weakDef);
+
+	// look in children that I re-export
+	for (const auto &dep : _dependentDylibs) {
+		if ( dep.reExport ) {
+			auto ret = dep.dylib->hasWeakDefinitionImpl(name);
+			if ( ret.first )
+				return ret;
+		}
+	}
+	return std::make_pair(false, false);
+}
+
+
 template <typename A>
 bool File<A>::hasWeakDefinition(const char* name) const
 {
 	// if supposed to ignore this export, then pretend I don't have it
 	if ( _ignoreExports.count(name) != 0 )
 		return false;
-		
-	typename NameToAtomMap::const_iterator pos = _atoms.find(name);
-	if ( pos != _atoms.end() ) {
-		return pos->second.weakDef;
-	}
-	else {
-		// look in children that I re-export
-		for (typename std::vector<Dependent>::const_iterator it = _dependentDylibs.begin(); it != _dependentDylibs.end(); ++it) {
-			if ( it->reExport ) {
-				//fprintf(stderr, "getJustInTimeAtomsFor: %s NOT found in %s, looking in child %s\n", name, this->path(), (*it)->getInstallPath());
-				typename NameToAtomMap::iterator cpos = it->dylib->_atoms.find(name);
-				if ( cpos != it->dylib->_atoms.end() ) 
-					return cpos->second.weakDef;
-			}
-		}
-	}
-	return false;
+
+	return hasWeakDefinitionImpl(name).second;
 }
 
 
@@ -665,24 +808,24 @@ bool File<A>::allSymbolsAreWeakImported() const
 
 
 template <typename A>
-bool File<A>::containsOrReExports(const char* name, bool* weakDef, bool* tlv, pint_t* defAddress) const
+bool File<A>::containsOrReExports(const char* name, bool& weakDef, bool& tlv, uint64_t& defAddress) const
 {
 	if ( _ignoreExports.count(name) != 0 )
 		return false;
 
-// check myself
-	typename NameToAtomMap::iterator pos = _atoms.find(name);
+	// check myself
+	const auto pos = _atoms.find(name);
 	if ( pos != _atoms.end() ) {
-		*weakDef = pos->second.weakDef;
-		*tlv = pos->second.tlv;
-		*defAddress = pos->second.address;
+		weakDef = pos->second.weakDef;
+		tlv = pos->second.tlv;
+		defAddress = pos->second.address;
 		return true;
 	}
 	
 	// check dylibs I re-export
-	for (typename std::vector<Dependent>::const_iterator it = _dependentDylibs.begin(); it != _dependentDylibs.end(); ++it) {
-		if ( it->reExport && !it->dylib->implicitlyLinked() ) {
-			if ( it->dylib->containsOrReExports(name, weakDef, tlv, defAddress) )
+	for (const auto &dep : _dependentDylibs) {
+		if ( dep.reExport && !dep.dylib->implicitlyLinked() ) {
+			if ( dep.dylib->containsOrReExports(name, weakDef, tlv, defAddress) )
 				return true;
 		}
 	}
@@ -700,7 +843,7 @@ bool File<A>::justInTimeforEachAtom(const char* name, ld::File::AtomHandler& han
 	
 	
 	AtomAndWeak bucket;
-	if ( this->containsOrReExports(name, &bucket.weakDef, &bucket.tlv, &bucket.address) ) {
+	if ( this->containsOrReExports(name, bucket.weakDef, bucket.tlv, bucket.address) ) {
 		bucket.atom = new ExportAtom<A>(*this, name, bucket.weakDef, bucket.tlv, bucket.address);
 		_atoms[name] = bucket;
 		_providedAtom = true;
@@ -747,6 +890,9 @@ bool File<A>::isPublicLocation(const char* pth)
 template <typename A>
 void File<A>::processIndirectLibraries(ld::dylib::File::DylibHandler* handler, bool addImplicitDylibs)
 {
+	// only do this once
+	if ( _indirectDylibsProcessed )
+		return;
 	const static bool log = false;
 	if ( log ) fprintf(stderr, "processIndirectLibraries(%s)\n", this->installPath());
 	if ( _linkingFlat ) {
@@ -764,7 +910,7 @@ void File<A>::processIndirectLibraries(ld::dylib::File::DylibHandler* handler, b
 				if ( log ) fprintf(stderr, "processIndirectLibraries() parent=%s, child=%s\n", this->installPath(), it->path);
 				// a LC_REEXPORT_DYLIB, LC_SUB_UMBRELLA or LC_SUB_LIBRARY says we re-export this child
 				it->dylib = (File<A>*)handler->findDylib(it->path, this->path());
-				if ( it->dylib->hasPublicInstallName() ) {
+				if ( it->dylib->hasPublicInstallName() && !it->dylib->wrongOS() ) {
 					// promote this child to be automatically added as a direct dependent if this already is
 					if ( (this->explicitlyLinked() || this->implicitlyLinked()) && (strcmp(it->path,it->dylib->installPath()) == 0) ) {
 						if ( log ) fprintf(stderr, "processIndirectLibraries() implicitly linking %s\n", it->dylib->installPath());
@@ -804,26 +950,28 @@ void File<A>::processIndirectLibraries(ld::dylib::File::DylibHandler* handler, b
 	chain.prev = NULL;
 	chain.file = this;
 	this->assertNoReExportCycles(&chain);
+	
+	_indirectDylibsProcessed = true;
 }
 
 template <typename A>
-void File<A>::assertNoReExportCycles(ReExportChain* prev)
+void File<A>::assertNoReExportCycles(ReExportChain* prev) const
 {
 	// recursively check my re-exported dylibs
 	ReExportChain chain;
 	chain.prev = prev;
 	chain.file = this;
-	for (typename std::vector<Dependent>::iterator it = _dependentDylibs.begin(); it != _dependentDylibs.end(); it++) {
-		if ( it->reExport ) {
-			ld::File* child = it->dylib;
+	for (const auto &dep : _dependentDylibs) {
+		if ( dep.reExport ) {
+			ld::File* child = dep.dylib;
 			// check child is not already in chain 
-			for (ReExportChain* p = prev; p != NULL; p = p->prev) {
+			for (ReExportChain* p = prev; p != nullptr; p = p->prev) {
 				if ( p->file == child ) {
 					throwf("cycle in dylib re-exports with %s and %s", child->path(), this->path());
 				}
 			}
-			if ( it->dylib != NULL )
-				it->dylib->assertNoReExportCycles(&chain);
+			if ( dep.dylib != nullptr )
+				dep.dylib->assertNoReExportCycles(&chain);
 		}
 	}
 }
@@ -836,79 +984,29 @@ public:
 	typedef typename A::P					P;
 
 	static bool										validFile(const uint8_t* fileContent, bool executableOrDyliborBundle);
+	static const char*								fileKind(const uint8_t* fileContent);
 	static ld::dylib::File*							parse(const uint8_t* fileContent, uint64_t fileLength, 
 															const char* path, time_t mTime, 
-															uint32_t ordinal, const Options& opts, bool indirectDylib) {
-															 return new File<A>(fileContent, fileLength, path, mTime,
+															ld::File::Ordinal ordinal, const Options& opts, bool indirectDylib) {
+															return new File<A>(fileContent, fileLength, path, mTime,
 																			ordinal, opts.flatNamespace(), 
 																			opts.linkingMainExecutable(),
 																			opts.implicitlyLinkIndirectPublicDylibs(), 
-																			opts.macosxVersionMin(), 
-																			opts.iOSVersionMin(),
+																			opts.platform(), 
+																			opts.minOSversion(),
+																			opts.allowSimulatorToLinkWithMacOSX(),
 																			opts.addVersionLoadCommand(),
-																			opts.logAllFiles(), 
+																			opts.targetIOSSimulator(),
+																			opts.logAllFiles(),
 																			opts.installPath(),
-																			indirectDylib);
+																			indirectDylib,
+																			opts.outputKind() == Options::kPreload);
 														}
 
 };
 
 
 
-template <>
-bool Parser<ppc>::validFile(const uint8_t* fileContent, bool executableOrDyliborBundle)
-{
-	const macho_header<P>* header = (const macho_header<P>*)fileContent;
-	if ( header->magic() != MH_MAGIC )
-		return false;
-	if ( header->cputype() != CPU_TYPE_POWERPC )
-		return false;
-	switch ( header->filetype() ) {
-		case MH_DYLIB:
-		case MH_DYLIB_STUB:
-			return true;
-		case MH_BUNDLE:
-			if ( executableOrDyliborBundle )
-				return true;
-			else
-				throw "can't link with bundle (MH_BUNDLE) only dylibs (MH_DYLIB)";
-		case MH_EXECUTE:
-			if ( executableOrDyliborBundle )
-				return true;
-			else
-				throw "can't link with a main executable";
-		default:
-			return false;
-	}
-}
-
-template <>
-bool Parser<ppc64>::validFile(const uint8_t* fileContent, bool executableOrDyliborBundle)
-{
-	const macho_header<P>* header = (const macho_header<P>*)fileContent;
-	if ( header->magic() != MH_MAGIC_64 )
-		return false;
-	if ( header->cputype() != CPU_TYPE_POWERPC64 )
-		return false;
-	switch ( header->filetype() ) {
-		case MH_DYLIB:
-		case MH_DYLIB_STUB:
-			return true;
-		case MH_BUNDLE:
-			if ( executableOrDyliborBundle )
-				return true;
-			else
-				throw "can't link with bundle (MH_BUNDLE) only dylibs (MH_DYLIB)";
-		case MH_EXECUTE:
-			if ( executableOrDyliborBundle )
-				return true;
-			else
-				throw "can't link with a main executable";
-		default:
-			return false;
-	}
-}
-
 template <>
 bool Parser<x86>::validFile(const uint8_t* fileContent, bool executableOrDyliborBundle)
 {
@@ -992,34 +1090,167 @@ bool Parser<arm>::validFile(const uint8_t* fileContent, bool executableOrDylibor
 
 
 
+template <>
+bool Parser<arm64>::validFile(const uint8_t* fileContent, bool executableOrDyliborBundle)
+{
+	const macho_header<P>* header = (const macho_header<P>*)fileContent;
+	if ( header->magic() != MH_MAGIC_64 )
+		return false;
+	if ( header->cputype() != CPU_TYPE_ARM64 )
+		return false;
+	switch ( header->filetype() ) {
+		case MH_DYLIB:
+		case MH_DYLIB_STUB:
+			return true;
+		case MH_BUNDLE:
+			if ( executableOrDyliborBundle )
+				return true;
+			else
+				throw "can't link with bundle (MH_BUNDLE) only dylibs (MH_DYLIB)";
+		case MH_EXECUTE:
+			if ( executableOrDyliborBundle )
+				return true;
+			else
+				throw "can't link with a main executable";
+		default:
+			return false;
+	}
+}
+
+
+bool isDylibFile(const uint8_t* fileContent, cpu_type_t* result, cpu_subtype_t* subResult)
+{
+	if ( Parser<x86_64>::validFile(fileContent, false) ) {
+		*result = CPU_TYPE_X86_64;
+		const macho_header<Pointer64<LittleEndian> >* header = (const macho_header<Pointer64<LittleEndian> >*)fileContent;
+		*subResult = header->cpusubtype();
+		return true;
+	}
+	if ( Parser<x86>::validFile(fileContent, false) ) {
+		*result = CPU_TYPE_I386;
+		*subResult = CPU_SUBTYPE_X86_ALL;
+		return true;
+	}
+	if ( Parser<arm>::validFile(fileContent, false) ) {
+		*result = CPU_TYPE_ARM;
+		const macho_header<Pointer32<LittleEndian> >* header = (const macho_header<Pointer32<LittleEndian> >*)fileContent;
+		*subResult = header->cpusubtype();
+		return true;
+	}
+	if ( Parser<arm64>::validFile(fileContent, false) ) {
+		*result = CPU_TYPE_ARM64;
+		*subResult = CPU_SUBTYPE_ARM64_ALL;
+		return true;
+	}
+	return false;
+}
+
+template <>
+const char* Parser<x86>::fileKind(const uint8_t* fileContent)
+{
+	const macho_header<P>* header = (const macho_header<P>*)fileContent;
+	if ( header->magic() != MH_MAGIC )
+		return NULL;
+	if ( header->cputype() != CPU_TYPE_I386 )
+		return NULL;
+	return "i386";
+}
+
+template <>
+const char* Parser<x86_64>::fileKind(const uint8_t* fileContent)
+{
+	const macho_header<P>* header = (const macho_header<P>*)fileContent;
+	if ( header->magic() != MH_MAGIC_64 )
+		return NULL;
+	if ( header->cputype() != CPU_TYPE_X86_64 )
+		return NULL;
+	return "x86_64";
+}
+
+template <>
+const char* Parser<arm>::fileKind(const uint8_t* fileContent)
+{
+	const macho_header<P>* header = (const macho_header<P>*)fileContent;
+	if ( header->magic() != MH_MAGIC )
+		return NULL;
+	if ( header->cputype() != CPU_TYPE_ARM )
+		return NULL;
+	for (const ArchInfo* t=archInfoArray; t->archName != NULL; ++t) {
+		if ( (t->cpuType == CPU_TYPE_ARM) && ((cpu_subtype_t)header->cpusubtype() == t->cpuSubType) ) {
+			return t->archName;
+		}
+	}
+	return "arm???";
+}
+
+#if SUPPORT_ARCH_arm64
+template <>
+const char* Parser<arm64>::fileKind(const uint8_t* fileContent)
+{
+  const macho_header<P>* header = (const macho_header<P>*)fileContent;
+  if ( header->magic() != MH_MAGIC_64 )
+    return NULL;
+  if ( header->cputype() != CPU_TYPE_ARM64 )
+    return NULL;
+  return "arm64";
+}
+#endif
+
+//
+// used by linker is error messages to describe mismatched files
+//
+const char* archName(const uint8_t* fileContent)
+{
+	if ( Parser<x86_64>::validFile(fileContent, true) ) {
+		return Parser<x86_64>::fileKind(fileContent);
+	}
+	if ( Parser<x86>::validFile(fileContent, true) ) {
+		return Parser<x86>::fileKind(fileContent);
+	}
+	if ( Parser<arm>::validFile(fileContent, true) ) {
+		return Parser<arm>::fileKind(fileContent);
+	}
+#if SUPPORT_ARCH_arm64
+	if ( Parser<arm64>::validFile(fileContent, false) ) {
+		return Parser<arm64>::fileKind(fileContent);
+	}
+#endif
+	return NULL;
+}
+
+
 //
 // main function used by linker to instantiate ld::Files
 //
 ld::dylib::File* parse(const uint8_t* fileContent, uint64_t fileLength, 
-							const char* path, time_t modTime, const Options& opts, uint32_t ordinal, 
+							const char* path, time_t modTime, const Options& opts, ld::File::Ordinal ordinal, 
 							bool bundleLoader, bool indirectDylib)
 {
 	switch ( opts.architecture() ) {
+#if SUPPORT_ARCH_x86_64
 		case CPU_TYPE_X86_64:
 			if ( Parser<x86_64>::validFile(fileContent, bundleLoader) )
 				return Parser<x86_64>::parse(fileContent, fileLength, path, modTime, ordinal, opts, indirectDylib);
 			break;
+#endif
+#if SUPPORT_ARCH_i386
 		case CPU_TYPE_I386:
 			if ( Parser<x86>::validFile(fileContent, bundleLoader) )
 				return Parser<x86>::parse(fileContent, fileLength, path, modTime, ordinal, opts, indirectDylib);
 			break;
+#endif
+#if SUPPORT_ARCH_arm_any
 		case CPU_TYPE_ARM:
 			if ( Parser<arm>::validFile(fileContent, bundleLoader) )
 				return Parser<arm>::parse(fileContent, fileLength, path, modTime, ordinal, opts, indirectDylib);
 			break;
-		case CPU_TYPE_POWERPC:
-			if ( Parser<ppc>::validFile(fileContent, bundleLoader) )
-				return Parser<ppc>::parse(fileContent, fileLength, path, modTime, ordinal, opts, indirectDylib);
-			break;
-		case CPU_TYPE_POWERPC64:
-			if ( Parser<ppc64>::validFile(fileContent, bundleLoader) )
-				return Parser<ppc64>::parse(fileContent, fileLength, path, modTime, ordinal, opts, indirectDylib);
+#endif
+#if SUPPORT_ARCH_arm64
+		case CPU_TYPE_ARM64:
+			if ( Parser<arm64>::validFile(fileContent, bundleLoader) )
+				return Parser<arm64>::parse(fileContent, fileLength, path, modTime, ordinal, opts, indirectDylib);
 			break;
+#endif
 	}
 	return NULL;
 }
