@@ -1242,12 +1242,15 @@ static ld::relocatable::File* createReader(const char* path)
 	objOpts.keepDwarfUnwind		= false;
 	objOpts.forceDwarfConversion = false;
 	objOpts.verboseOptimizationHints = true;
+	objOpts.armUsesZeroCostExceptions = true;
 	objOpts.subType				= sPreferredSubArch;
+	objOpts.srcKind				= ld::relocatable::File::kSourceObj;
 #if 1
 	if ( ! foundFatSlice ) {
 		cpu_type_t archOfObj;
 		cpu_subtype_t subArchOfObj;
-		if ( mach_o::relocatable::isObjectFile(p, &archOfObj, &subArchOfObj) ) {
+		Options::Platform platform;
+		if ( mach_o::relocatable::isObjectFile(p, &archOfObj, &subArchOfObj, &platform) ) {
 			objOpts.architecture = archOfObj;
 			objOpts.subType = subArchOfObj;
 		}
