@@ -1,8 +1,10 @@
 #include <stdio.h>
 
+extern void back();
 
 void foo()
 {
 	fprintf(stdout, "foo\n");
+  back();
 }
 
